#####################################################
#  Module Pastel::Graphics.pm
#  Graphics context class of SVG document. 
#  
#  Author: Malay
#  Email : curiouser@ccmb.res.in
######################################################

package Pastel::Graphics;


#use Pastel::Mixin::Mixin; # For _rearrange function
#use Pastel::BasicStroke;
#use Pastel::Geometry::Shape;
#use Pastel::Geometry::Point;
#use Pastel::Geometry::Line;
#use Pastel::Geometry::Rectangle;
#use Pastel::Color;
#use Pastel::Tools::Writer;
#use Pastel::Font;
#use Pastel::Text;
#use Pastel::Geometry::Ellipse;

@ISA = qw( Pastel::Mixin::Mixin);
use strict;
use Carp;


my $_width    =   "800"; #Width of the SVG document
my $_height   =   "600"; #Height of the SVG document
my $_dtd      =    '<!DOCTYPE svg PUBLIC "-//W3C//DTD SVG 20001102//EN" "http://www.w3.org/TR/2000/CR-SVG-20001102/DTD/svg-20001102.dtd">';


sub new
{
    my $arg = shift;
    my $class = ref($arg) || $arg;
    my $self = {};
    bless $self, $class;
   
    $self->_init(@_);

    return $self;

}

sub _init {
   
    my ($self, @args) = @_;
   
    my ($w, $h,  $dtd, $cgi) = $self->_rearrange(
					  ["WIDTH","HEIGHT","DTD","CGI"],
					   @args);

    $self->{width} = defined ($w) ?  $w  : $_width; 
    $self->{height} = defined ($h)? $h : $_height;
    $self->{dtd} = defined ($dtd) ? $dtd : $_dtd;
    $self->{CGI} = defined ($cgi) ? $cgi : "true";
    $self->{_temp} = ""; # To store SVG elements
    $self->{_defs} = ""; # To store DEFS
    $self->{_writer} = $self->_initialize_writer();
    $self->{paint} = undef;
    $self->{stroke} = undef;
    $self->{defs} = [];
}

sub _buffer_add {
    my $self = shift;
    my $s = shift;
    $self->{_temp} .= $s;
}

sub _initialize_writer {
    my $self = shift;
    my $writer = Pastel::Tools::Writer->new();
    return $writer;
}

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


sub set_font {
    my ($self, $arg) =  @_;
     $self->{"font"} = $arg;
    return $self;
    #   my ($self,%arg) = @_;
#    if(defined ($arg{FAMILY})){$self->{FONT_FAMILY} = $arg{FAMILY};}
#    if(defined ($arg{SIZE})){$self->{FONT_SIZE} = $arg{SIZE};}
#    if(defined($arg{COLOR})){$self->{FONT_COLOR} = $arg{COLOR};}
#    if(defined($arg{WEIGHT})){$self->{FONT_WEIGHT} = $arg{WEIGHT};}
#    return $self;
}

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

sub set_fill_color {
    my $self = shift;
    $self->{COLOR} = shift;
    return $self;
}

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


sub set_stroke {
    my $self = shift;
    my $arg = shift;
    #print "******STROKE:", $arg->_to_svg(),"\n";
    if ( defined($arg)  ){
	#print "Inside CLEAR\n";
	$self->{stroke} = $arg;
	}else{
	if (defined($self->{stroke}) ){
	    $self->{stroke} = undef;
	    return;
	}
    }

}

sub get_stroke {
  my $self = shift;
  if(!defined ($self->{stroke})){
    $self->{stroke} = Pastel::BasicStroke->new(-width=>1);
    return $self->{stroke};
  }else{
  return $self->{stroke};
   }
}



#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Title:     draw_rect()
# Usage:     draw_rect(-x=>"n", -y=>"n", -width=>"n", height=>"n");
# Function:  Draws a open rectangle starting from co-ordinate (x,y) with
#            specified width and height. Use fillrect() to draw rectangle
#            with fill
# Args:      Coordinate positions x,y
#            Width in pixels
#            Height in pixels
#            Inherited properties: stroke-color and stroke-height


sub draw_rect {

  # Drawing logic: First check whether stroke is defined or not. 
  # if it is defined then draw it with the 
    my $self = shift;
    my ($x, $y, $width, $height)= $self->_rearrange(
                                  ["X","Y","WIDTH","HEIGHT"], @_);
    my $color = $self->get_paint()->to_svg_stroke();
    my $stroke = $self->get_stroke()->to_svg();
    $self->{_temp} .= ( qq(<rect x="$x" y="$y" width="$width" height="$height" style="fill:none; $color $stroke" />));
                     
}

=head2 fill_rect()

 Title    : fill_rect()
 Usage    : fill_rect(-x=>"n", -y=>"n", -width=>"n", -height=>"n");
 Function : Draws a filled rectangle starting from co-ordinate (x,y) with
            specified width and height.
 Args     : Coordinate positions x,y
            Width in pixels
            Height in pixels
 Returns  : Nothing

=cut

sub fill_rect {
    my $self = shift;
    my $color = $self->get_paint()->to_svg_fill();
    my ($x, $y, $width, $height)= $self->_rearrange(
                                  ["X","Y","WIDTH","HEIGHT"], @_);
    $self->{_temp} .= ( qq(<rect x="$x" y="$y" width="$width" height="$height" style="$color stroke:none"/>));
                     
}


=head2 draw_round_rect()

 Usage   : $g->draw_round_rect($x, $y, $width, $height, $arc_width, $arc_height);
 Args    : $x, $y are the top left hand corner. $width and $height are width and
           height. $arc_width- horizontal diameter of the arc at the four 
           corners. $arc_height - vertical diameter of the arc at the four corners
 Function: Draws a round cornered rectangle using the current "paint".
 Returns : Nothing

=cut

sub draw_round_rect {
    my ($self,@args) = @_;
    
    if ( @args < 6) {
      croak 'Missing parameter in Pastel::Graphics::draw_round_rect()!\n';
    }
    
    my $paint = $self->get_paint()->to_svg_stroke();
    my $stroke = $self->get_stroke()->to_svg();

    $self->{_temp} .= qq(<rect x="$args[0]" y="$args[1]" width="$args[2]" height="$args[3]" rx="$args[4]" ry="$args[4]" style="fill:none; $paint $stroke" />);

}

=head2 draw_fill_round_rect()

 Usage   : $g->draw_round_rect($x, $y, $width, $height, $arc_width, $arc_height);
 Args    : $x, $y are the top left hand corner. $width and $height are width and
           height. $arc_width- horizontal diameter of the arc at the four 
           corners. $arc_height - vertical diameter of the arc at the four corners
 Function: Draws a round cornered filled rectangle using the current "paint".
 Returns : Nothing

=cut

sub draw_fill_round_rect {
    my ($self,@args) = @_;
    
    if ( @args < 6) {
      croak 'Missing parameter in Pastel::Graphics::draw_round_rect()!\n';
    }
    
    my $paint = $self->get_paint()->to_svg_fill();
   # my $stroke = $self->get_stroke()->to_svg();

    $self->{_temp} .= qq(<rect x="$args[0]" y="$args[1]" width="$args[2]" height="$args[3]" rx="$args[4]" ry="$args[4]" style="$paint stroke=none;" />);

}

=head2 draw_3D_rect ()

  Usage   : $g->draw_3D_rect($x, $y, $width, $height, $raised);
  Args    : $x - Top left corner X coordinate  of the rectangle
            $y - Top left corner Y coordinate  of the rectangle
            $width - Width of the rectangle
            $height - Height of the rectangle
            $raised - take value "true" or "false"
  Function: The function draws 3D effect rectangle. If stroke is set then the 
            stroke width is taken from the "stroke" object.
  Returns : Nothing

=cut

sub draw_3D_rect {
    my ($self, @args) = @_;
    if ( @args < 5) {
       croak 'Missing parameter in Pastel::Graphics::draw_3D_rect()!\n';
    }
    my $raised;
    if ( $args[4] =~ /true/i){
      $raised = 1;
    }else {
      $raised = 0;
    }
    my $color = $self->get_paint()->to_svg_stroke();
    my $width = $self->get_stroke()->get_line_width();
    my $stroke = $self->get_stroke();
     $self->set_stroke(Pastel::BasicStroke->new(-width=>1));
    if( $raised ){
       $self->set_paint($self->get_paint()->brighter());
       $self->draw_line($args[0], $args[1], $args[0],$args[1] + $args[3]);
       $self->draw_line($args[0]+ 1, $args[1], $args[0]+ $args[2]- 1,$args[1]);
       $self->set_paint($self->get_paint()->darker());
       $self->draw_line($args[0]+ 1, $args[1]+$args[3], $args[0]+$args[2],$args[1] + $args[3]);
       $self->draw_line($args[0]+$args[2], $args[1], $args[0]+ $args[2], $args[1] + $args[3]- 1);
    }else {
       # = $color->darker();
    }
   $self->set_paint($color);
   $self->set_stroke($stroke);
    
    

}
sub _print_header {
	my $self = shift;
	$self->{_writer}->print_svg(
			       '<?xml version="1.0" encoding="iso-8859-1"?>');
	$self->{_writer}->print_svg($self->{dtd});
}

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

sub _print_footer {
	my $self = shift;
 	$self->{_writer}->print_svg("</svg>");
}

sub _print_svg_start {
    my $self = shift;
    $self->{_writer}->print_svg("<svg width=\"".$self->{width}."\" height=\"".$self->{height}."\">" );
}


sub message_err {
    my $self    = shift;
    my $message = shift;
    if (($self->{DEBUG} =~/^true$/i)&&($self->{CGI} ne "true")){
       my $fh = $self->{ERROR_HANDLE};
       print $fh "ERROR: $message\n";
    }
}


sub set_paint {
  my $self =shift;
  $self->{paint} = $_[0];
}

sub get_paint {
  my $self = shift;
  if(!defined($self->{paint})){
    $self->{paint} = Pastel::Color->black();
  }
  return $self->{paint};
}

sub draw {
  my $self = shift;
  my $ob = shift;
  my $data;
  
  
  #if (defined ($self->{stroke}) ){
   #print "***********Stroke defined********\n";
    #my $stroke =  $self->{stroke};
    #my $stroked_ob = $stroke->create_stroked_shape($ob);
     #$data = $stroked_$ob->_draw($self);
  #}
  #else{
      $data = $ob->_draw($self);
  #}

  $self->_buffer_add($data);
}

=head2

Title :   draw_string( Pastel::Text::$string, int/float $x, int/float $y)
          draw_string( "perl string", int/float $x, int/float $y)
          draw_string(-text=>$string, -x=>220, -y=>440);

Usage :   my $s = Pastel::String->new("Fantastic");
          $g->draw_string($s, 200, 300);

          # same thing
          $g->draw_string(-string=>$s, -x=>200, -y=>300); 

          # same thing
          $g->draw_string("Fantastic", 200, 300);

Function: Draws the string supplied in the co-ordinates spefied. The font
          of the string will be determined the font specifed of the
          graphics object with the set_font() method.

Args:     The method takes 3 arguments. The first argument can be an object
          of type Pastel::String or a pure perl string. The string supplied is
          converted to XML form automatically. All character code above 126
          is converted to "&xff;" form. "<", ">", "&" and other XML entities
          are automatically created. Please note Pastel::Graphics can handle only
          ISO-8859 charcter sets. Unicode is not yet implemented.

=cut

sub draw_string
{
    my ($self, @args) = @_;
    my ($string, $x, $y) = $self->_rearrange(["TEXT","X","Y"], @args);

    my $data =""; #Holds the SVG string;
    
    unless ( (defined ($string)) && (defined($x)) && (defined($y)) ){
	print 'Error: Graphics::drawstring(): Argument missing', "\n";
	return;
    }

    # If $string is a string object calls its draw method
    if( ref($string) =~ m/Text/ ){

	$string->set_x($x);
	$string->set_y($y);
	
        if( defined ($self->{font} )){ # If the graphics object has a font
	    # First fontify it
	    $string->set_font( $self->get_font() );
	    $data = $string->draw($self);

	} else{
	    $data = $string->draw($self);
	}
    }

    # If $string is a perl string, create a string object, fontify it
    # by passing it to the font object and then call its draw method
    else {
	my $string = Pastel::Text->new($string);
	$string->set_x($x);
	$string->set_y($y);

	if ( defined($self->{font}) ){  # If the graphics object has font

            # Set the font object of the String
	    my $fontified = $string->set_font($self->{font});
	    $data = $fontified->draw($self);
	}
	else { # If the graphics object doesn't have font object set
	    $data = $string->draw($self); # Just call the draw method
	}
    }

    $self->_buffer_add($data); # Add the data to the print buffer

}
sub show {
    my $self = shift;
    my $writer = $self->{_writer};
    $self->_print_header();
    $self->_print_svg_start();
    
    if( @{ $self->{defs} } > 0 ){
	$writer->print_svg('<defs>');
	
	for(my $i = 0; $i < @{ $self->{defs} }; $i++){
	    $writer->print_svg( ${$self->{defs}}[$i]->to_svg() );
    }
    $writer->print_svg('</defs>');
}
    

	
    $writer->print_svg( $self->{_temp} );
    $self->_print_footer();
}

sub get_font {
    my $self = shift;
    if ( defined ( $self->{font} ) ) {
	return $self->{font};
    } else {
	return undef;
	}
}

sub fill {
    my $self = shift;
    my $shape = shift;
    
    if(!$shape->isa("Pastel::Shape")){
	croak 'Error: Only a shape can be filled by Pastel::Graphics->fill()';
    }
    
    $shape->set_fill($self->get_paint());
}

sub add_to_defs {
    my $self = shift;
    my $object = shift;
    ${$self->{defs}}[@{$self->{defs}}] = $object;
}

sub get_svg {
  my $self = shift;
  my $s =  '<?xml version="1.0" encoding="iso-8859-1"?>';
   $s .= "\n".$self->{dtd}."\n";
   $s .= "<svg width=\"".$self->{width}."\" height=\"".$self->{height}."\">";
  if( @{ $self->{defs} } > 0 ){
    $s .="\n".'<defs>'."\n";
    for(my $i = 0; $i < @{ $self->{defs} }; $i++){
	    $s .= ${$self->{defs}}[$i]->to_svg();
  }
    $s .= "\n".'</defs>'."\n";
}
   $s .= $self->{_temp}; 
   $s .= "\n".'</svg>';
   return $s;
}

sub draw_polyline {
  my @args = @_;
  my $self = $args[0];
  my @x = @{$args[1]};
  my @y = @{$args[2]};

  my $s = '<polyline style="fill:none; ';
   $s .= $self->get_paint()->to_svg();
   $s .= " ".$self->get_stroke()->to_svg()."\" ";
   $s .= 'points="';
  
  for ( my $i=0; $i < @x; $i++){
    $s .= $x[$i].",".$y[$i];
    if( $i != (scalar(@x) - 1)){
      $s .=" ";
    }
  }
=head1
  my $s .= '<path style="fill:none; ';
  $s .= $self->get_paint()->to_svg();
  $s .= " ".$self->get_stroke()->to_svg()."\" ";
  $s .= 'd="M'.$x[0].",".$y[0]." ";
  for ( my $i=0; $i < @x; $i++){
    $s .= "L".$x[$i].",".$y[$i];
    if( $i != (scalar(@x) - 1)){
      $s .=" ";
    }
  }
=cut
  $s .='" />';
  $self->{_temp} .= $s."\n";
}  
  
sub draw_line {
  my ($self, @args) = @_;
  if ( @args < 4){
    croak "Missing parameters in Pastel::Graphics::draw_line()!\n";
  }
  my $color = $self->get_paint()->to_svg_stroke();
  my $stroke = $self->get_stroke()->to_svg();
  $self->{_temp} .=  qq(<line x1="$args[0]" y1="$args[1]" x2="$args[2]" y2="$args[3]" style="fill:none; $color $stroke" />);
}

sub draw_oval {
  my ($self, @args) = @_;
  if ( @args < 4){
    croak "Missing parameters in Pastel::Graphics::draw_line()!\n";
  }
  my $color = $self->get_paint()->to_svg_stroke();
  my $stroke = $self->get_stroke()->to_svg();

  my $cx = $args[0] + ( $args[2]/2);
  my $cy = $args[1] + ( $args[3]/2);
  my $rx = $args[2]/2;
  my $ry = $args[3]/2;
  $self->{_temp} .=  qq(<ellipse cx="$cx" cy="$cy" rx="$rx" ry="$ry" style="fill:none; $color $stroke" />);
}

sub fill_oval {
  my ($self, @args) = @_;
  if ( @args < 4){
    croak "Missing parameters in Pastel::Graphics::draw_line()!\n";
  }
  my $color = $self->get_paint()->to_svg_fill();
  my $stroke = $self->get_stroke()->to_svg();

  my $cx = $args[0] + ( $args[2]/2);
  my $cy = $args[1] + ( $args[3]/2);
  my $rx = $args[2]/2;
  my $ry = $args[3]/2;
  $self->{_temp} .=  qq(<ellipse cx="$cx" cy="$cy" rx="$rx" ry="$ry" style="$color $stroke" />);
}

sub get_height {
    return $_[0]->{height};
}

sub set_height {
  if ($_[1]){ $_[0]->{height} = $_[1];}
}

sub get_width {
  return $_[0]->{width};
}

sub set_width {
  if ($_[1]){ $_[0]->{width} = $_[1];}
}
1;


